package sun.nio.fs;

import java.nio.file.NotLinkException;
import java.io.IOError;
import java.nio.file.FileSystemException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.IOException;
import sun.misc.Unsafe;

class WindowsLinkSupport
{
    private static final Unsafe unsafe;
    
    private WindowsLinkSupport() {
    }
    
    static String readLink(WindowsPath windowsPath) throws IOException {
        long openForReadAttributeAccess = 0L;
        try {
            openForReadAttributeAccess = windowsPath.openForReadAttributeAccess(false);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(windowsPath);
        }
        try {
            return readLinkImpl(openForReadAttributeAccess);
        }
        finally {
            WindowsNativeDispatcher.CloseHandle(openForReadAttributeAccess);
        }
    }
    
    static String getFinalPath(WindowsPath windowsPath) throws IOException {
        long openForReadAttributeAccess = 0L;
        try {
            openForReadAttributeAccess = windowsPath.openForReadAttributeAccess(true);
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(windowsPath);
        }
        try {
            return stripPrefix(WindowsNativeDispatcher.GetFinalPathNameByHandle(openForReadAttributeAccess));
        }
        catch (WindowsException ex2) {
            if (ex2.lastError() != 124) {
                ex2.rethrowAsIOException(windowsPath);
            }
        }
        finally {
            WindowsNativeDispatcher.CloseHandle(openForReadAttributeAccess);
        }
        return null;
    }
    
    static String getFinalPath(final WindowsPath resolve, boolean b) throws IOException {
        WindowsFileSystem fileSystem = resolve.getFileSystem();
        try {
            if (!b || !fileSystem.supportsLinks()) {
                return resolve.getPathForWin32Calls();
            }
            if (!WindowsFileAttributes.get(resolve, false).isSymbolicLink()) {
                return resolve.getPathForWin32Calls();
            }
        }
        catch (WindowsException ex) {
            ex.rethrowAsIOException(resolve);
        }
        String finalPath = getFinalPath(resolve);
        if (finalPath != null) {
            return finalPath;
        }
        int n = 0;
        while (true) {
            try {
                if (!WindowsFileAttributes.get(resolve, false).isSymbolicLink()) {
                    return resolve.getPathForWin32Calls();
                }
            }
            catch (WindowsException ex2) {
                ex2.rethrowAsIOException(resolve);
            }
            WindowsPath fromNormalizedPath = WindowsPath.createFromNormalizedPath(fileSystem, readLink(resolve));
            WindowsPath windowsPath = resolve.getParent();
            if (windowsPath == null) {
                windowsPath = ((WindowsPath)AccessController.doPrivileged(new PrivilegedAction<WindowsPath>() {
                    @Override
                    public WindowsPath run() {
                        return resolve.toAbsolutePath();
                    }
                })).getParent();
            }
            resolve = windowsPath.resolve(fromNormalizedPath);
            if (++n >= 32) {
                break;
            }
            continue;
        }
        throw new FileSystemException(resolve.getPathForExceptionMessage(), null, "Too many links");
    }
    
    static String getRealPath(WindowsPath windowsPath, boolean b) throws IOException {
        WindowsFileSystem fileSystem = windowsPath.getFileSystem();
        if (b && !fileSystem.supportsLinks()) {
            b = false;
        }
        String s;
        try {
            s = windowsPath.toAbsolutePath().toString();
        }
        catch (IOError ioError) {
            throw (IOException)(IOException)ioError.getCause();
        }
        if (s.indexOf(46) >= 0) {
            try {
                s = WindowsNativeDispatcher.GetFullPathName(s);
            }
            catch (WindowsException ex) {
                ex.rethrowAsIOException(windowsPath);
            }
        }
        StringBuilder o = new StringBuilder(s.length());
        int char1 = s.charAt(0);
        int char2 = s.charAt(1);
        int i;
        if (((char1 <= 'z' && char1 >= 'a') || (char1 <= 'Z' && char1 >= 'A')) && char2 == ':' && s.charAt(2) == '\\') {
            o.append(Character.toUpperCase((char)char1));
            o.append(":\\");
            i = 3;
        }
        else if (char1 == '\\' && char2 == '\\') {
            int n = s.length() - 1;
            int index = s.indexOf(92, 2);
            if (index == -1 || index == n) {
                throw new FileSystemException(windowsPath.getPathForExceptionMessage(), null, "UNC has invalid share");
            }
            int index2 = s.indexOf(92, index + 1);
            if (index2 < 0) {
                index2 = n;
                o.append(s).append("\\");
            }
            else {
                o.append(s, 0, index2 + 1);
            }
            i = index2 + 1;
        }
        else {
            throw new AssertionError("path type not recognized");
        }
        if (i >= s.length()) {
            String string2 = o.toString();
            try {
                WindowsNativeDispatcher.GetFileAttributes(string2);
            }
            catch (WindowsException ex3) {
                ex3.rethrowAsIOException(s);
            }
            return string2;
        }
        int index3;
        for (; i < s.length(); i = index3 + 1) {
            index3 = s.indexOf(92, i);
            index3 = ((index3 == -1) ? s.length() : index3);
            String string = o.toString() + s.substring(i, index3);
            try {
                WindowsNativeDispatcher.FirstFile findFirstFile = WindowsNativeDispatcher.FindFirstFile(WindowsPath.addPrefixIfNeeded(string));
                WindowsNativeDispatcher.FindClose(findFirstFile.handle());
                if (b && WindowsFileAttributes.isReparsePoint(findFirstFile.attributes())) {
                    String s2 = getFinalPath(windowsPath);
                    if (s2 == null) {
                        s2 = getRealPath(resolveAllLinks(WindowsPath.createFromNormalizedPath(fileSystem, s)), false);
                        return s2;
                    }
                    return s2;
                }
                else {
                    o.append(findFirstFile.name());
                    if (index3 != -1) {
                        o.append('\\');
                    }
                }
            }
            catch (WindowsException ex2) {
                ex2.rethrowAsIOException(s);
            }
        }
        return o.toString();
    }
    
    private static String readLinkImpl(long n) throws IOException {
        int n2 = 16384;
        NativeBuffer nativeBuffer = NativeBuffers.getNativeBuffer(n2);
        try {
            try {
                WindowsNativeDispatcher.DeviceIoControlGetReparsePoint(n, nativeBuffer.address(), n2);
            }
            catch (WindowsException ex) {
                if (ex.lastError() == 4390) {
                    throw new NotLinkException(null, null, ex.errorString());
                }
                ex.rethrowAsIOException((String)null);
            }
            if ((int)WindowsLinkSupport.unsafe.getLong(nativeBuffer.address() + 0L) != -1610612724) {
                throw new NotLinkException(null, null, "Reparse point is not a symbolic link");
            }
            int short1 = WindowsLinkSupport.unsafe.getShort(nativeBuffer.address() + 8L);
            int short2 = WindowsLinkSupport.unsafe.getShort(nativeBuffer.address() + 10L);
            if (short2 % 2 != 0) {
                throw new FileSystemException(null, null, "Symbolic link corrupted");
            }
            char[] array = new char[short2 / 2];
            WindowsLinkSupport.unsafe.copyMemory(null, nativeBuffer.address() + 20L + (long)short1, array, (long)Unsafe.ARRAY_CHAR_BASE_OFFSET, (long)short2);
            String stripPrefix = stripPrefix(new String(array));
            if (stripPrefix.length() == 0) {
                throw new IOException("Symbolic link target is invalid");
            }
            return stripPrefix;
        }
        finally {
            nativeBuffer.release();
        }
    }
    
    private static WindowsPath resolveAllLinks(WindowsPath windowsPath) throws IOException {
        assert !(!windowsPath.isAbsolute());
        WindowsFileSystem fileSystem = windowsPath.getFileSystem();
        int n = 0;
        int i = 0;
        while (i < windowsPath.getNameCount()) {
            WindowsPath resolve = windowsPath.getRoot().resolve(windowsPath.subpath(0, i + 1));
            WindowsFileAttributes value = null;
            try {
                value = WindowsFileAttributes.get(resolve, false);
            }
            catch (WindowsException ex) {
                ex.rethrowAsIOException(resolve);
            }
            if (value.isSymbolicLink()) {
                if (++n > 32) {
                    throw new IOException("Too many links");
                }
                WindowsPath fromNormalizedPath = WindowsPath.createFromNormalizedPath(fileSystem, readLink(resolve));
                WindowsPath subpath = null;
                int nameCount = windowsPath.getNameCount();
                if (i + 1 < nameCount) {
                    subpath = windowsPath.subpath(i + 1, nameCount);
                }
                windowsPath = resolve.getParent().resolve(fromNormalizedPath);
                try {
                    String getFullPathName = WindowsNativeDispatcher.GetFullPathName(windowsPath.toString());
                    if (!getFullPathName.equals(windowsPath.toString())) {
                        windowsPath = WindowsPath.createFromNormalizedPath(fileSystem, getFullPathName);
                    }
                }
                catch (WindowsException ex2) {
                    ex2.rethrowAsIOException(windowsPath);
                }
                if (subpath != null) {
                    windowsPath = windowsPath.resolve(subpath);
                }
                i = 0;
            }
            else {
                ++i;
            }
        }
        return windowsPath;
    }
    
    private static String stripPrefix(String s) {
        if (s.startsWith("\\\\?\\")) {
            if (s.startsWith("\\\\?\\UNC\\")) {
                s = "\\" + s.substring(7);
                return s;
            }
            s = s.substring(4);
            return s;
        }
        else if (s.startsWith("\\??\\")) {
            if (s.startsWith("\\??\\UNC\\")) {
                s = "\\" + s.substring(7);
                return s;
            }
            s = s.substring(4);
            return s;
        }
        else {
            return s;
        }
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
    }
}
